#!/usr/bin/perl
# NDMP Index File Parser
# Takes a backup number and an NDMP index file as arguments and adds the files
# and directories therein to the FILES_TABLE and DIRS_TABLE, respectively. If
# no index file is provided, an attempt will be made to mount the backup,
# extract the index file, and proceed.
#
# The verbose (-v) and debug (-d) options send their output to STDERR. If
# calling this script from elsewhere, consider using the -v option and
# redirecting STDERR to append to <LOG_DIR>/ndmp_index_parser.log.
#
# This script returns 0 on success and the error code returned by the DBD::Pg
# module if an error occurs. If any file names were not able to be placed in
# the database (typically due to character encoding problems), the script
# returns 254. This is not, strictly speaking, a failure, and this exit code
# should be checked for when calling.
#
# NOTE CentOS 5 and CentOS 6 have different versions of the DBD::Pg module. The
# CentOS 5 version is heavily deprecated and requires slightly different usage
# than the CentOS 6 version. The NEW_DBD_VERS constant is used to differentiate
# between the two usages.

use strict;
use warnings;
use DBI;
use DBD::Pg;
use URI::Escape;
use Getopt::Std;

use constant DATABASE => "bpdb";
use constant DB_USER => "postgres";
use constant DIRS_TABLE => "bp.dirs_non_ctar";
use constant FILES_TABLE => "bp.backup_files_non_ctar";
use constant END_BACKUP_NO => 2147483647; # Convention for non-chained backups
use constant FT_REG => 0; # Internal file type identifier for regular files
use constant FT_DIR => 53; # Internal file type identifier for directories
use constant FT_CONTAINER => 57; # Internal file type identifier for containers
use constant NDMP_MOUNT_PATH => "/tmp/ndmp"; # Temporary location to mount NDMP backups
use constant VOLUME => '@@@:'; # Special volume container entry
use constant NEW_DBD_VERS => 2.0; # Earliest known version to use new functions

our (%opts, $backup_no, $index_file, %lookup, $mounted);

my $dbh;   # Database handler
my $key;   # Key for hash-map
my $query; # Query string
my $sth;   # Statement handler
my $sth2;  # Statement handler

# Print usage and exit
sub print_usage {
	print STDERR "This script parses NDMP index files and adds the names ",
		"and file attributes of files and directories in the backup to ",
		"a database.\n",
		"Usage: $0 [-h] [-v] [-d] -b <BACKUP_NUMBER> [-f <FILE>]\n",
		"-h\tPrint this help and exit.\n",
		"-v\tVerbose output.\n",
		"-d\tDebugging output (even more verbose).\n",
		"-b NUM\tBackup number.\n",
		"-f FILE\tNDMP index file. If not provided, an attempt will be made to ",
		"extract it based on the backup number.\n";
	exit;
}

# Connect to the database
sub db_connect {
	print STDERR "Connecting to database...\n" if ($opts{v} or $opts{d});
	$dbh = DBI->connect("dbi:Pg:dbname=" . DATABASE, DB_USER, "")
		or die "Could not connect to database server\n";
}

# Disconnect from the database
sub db_disconnect {
	print STDERR "Disconnect from database\n" if ($opts{v} or $opts{d});
	print STDERR "$_[0]" if ($_[0] and ($opts{v} or $opts{d})); # Optional message
	$dbh->disconnect();
}

# Mount backup
sub mount {
	print STDERR "Mount backup $backup_no\n" if ($opts{v} or $opts{d});
	system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "mount",
		$backup_no) == 0 or die "Failed to mount backup $backup_no\n";
	$index_file = NDMP_MOUNT_PATH . "/_dump/$backup_no/index";
	$mounted = 1;
}

# Unmount backup if $mounted
sub unmount {
	if ($mounted) {
		print STDERR "Unmount backup $backup_no\n" if ($opts{v} or $opts{d});
		system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "destroy",
			$backup_no) == 0 or die "Failed to unmount backup $backup_no\n";
	}
}

# Prepare to exit, especially in case of an error. Call before dying.
sub cleanup {
	&db_disconnect;
	&unmount;
}

# Initialize the script (read options, connect to database)
sub init {
	my $opt_string = 'hvdf:b:';
	getopts("$opt_string", \%opts) or print_usage();
	print_usage() if $opts{h} or not %opts; # exits
	$index_file = $opts{f};
	$backup_no = $opts{b};
	$mounted = 0;
	&db_connect;
}

&init;

# Unless an index file was specified, and exists, attempt to mount the backup
# and use the index file contained within
&mount unless ($opts{f} and -e $opts{f});

if ($opts{v} or $opts{d}) {
	print STDERR "Index File: $index_file\n",
		"Backup No.: $backup_no\n",
}

# Prepare INSERT/SELECT statements for DIRS_TABLE
$sth = $dbh->prepare("INSERT INTO " . DIRS_TABLE . " (dname) SELECT ? " .
	"WHERE NOT EXISTS (SELECT TRUE FROM " . DIRS_TABLE . " WHERE dname = ?)");
$sth2 = $dbh->prepare("SELECT dir_no FROM " . DIRS_TABLE .  " WHERE dname = ?");

# The blank directory entry should be added by the schema definition, but is
# included here for safety and consistencty.
$sth->execute('', '');
$sth2->execute('');
my $volume_dir_no = $sth2->fetchrow_array();

# Give cmc_dump_mount time to finish mounting. Using sleep is not ideal, but
# all the implicit forking and asynchronous commands means waiting for signals
# can potentially hang indefinitely; this approach allows us to set a timeout.
my $time = 1800;
while ($mounted and not -e $index_file and $time) {
	print STDERR "Mounting backup $backup_no, $time seconds remaining\n" if ($opts{d} or $opts{v});
	$time -= 10;
	sleep 10;
}

if (open(NDMP_INDEX, "<", $index_file)) {
	print STDERR "" if ($opts{v} or $opts{d});
	print STDERR "Parse $index_file... " if ($opts{v} or $opts{d});

	# Read through the index file and add each entry to the lookup hash table.
	while (<NDMP_INDEX>) {
		chomp;
		# Type - Parent - Name - Host Type? - Node
		if ($_ =~ /^DHd (\d+) (.+) UNIX (\d+)/) {
			# Node
			if ($2 ne '.' and $2 ne '..') {
				$lookup{$3}{'name'} = uri_unescape($2);
				$lookup{$3}{'parent'} = $1;
				$lookup{$3}{'mod_date'} = 0;
				$lookup{$3}{'block_offset'} = 0;
				$lookup{$3}{'type'} = FT_REG;
			}
		} elsif ($_ =~ /^DHn (\d+) UNIX fd m(\d+) u(\d+) g(\d+) tm(\d+) @(\d+)/) {
			# Directory
			if (exists $lookup{$1}) {
				$lookup{$1}{'mod_date'} = $5;
				$lookup{$1}{'block_offset'} = $6;
				$lookup{$1}{'type'} = FT_DIR;
				$lookup{$3}{'dir_no'} = $volume_dir_no;
			}
		} elsif ($_ =~ /^DHn (\d+) UNIX f- m(\d+) u(\d+) g(\d+) s(\d+) tm(\d+) @(\d+)/) {
			# File
			if (exists $lookup{$1}) {
				$lookup{$1}{'mod_date'} = $6;
				$lookup{$1}{'block_offset'} = $7;
			}
		} elsif ($_ =~ /^DHr (\d+)$/) {
			# Root
			unless (exists $lookup{$1}) {
				$lookup{$1}{'name'} = '/';
				$lookup{$1}{'parent'} = 0;
				$lookup{$1}{'mod_date'} = 0;
				$lookup{$1}{'block_offset'} = 0;
				$lookup{$1}{'type'} = FT_DIR;
			}
		} else {
			# No matches
			print "$_\n" if ($opts{d});
		}
	}

	print STDERR "Done\n" if ($opts{v} or $opts{d});
	close(NDMP_INDEX);
} else {
	&cleanup;
	die "Cannot open '$index_file': $!\n";
}

&unmount;

print STDERR "Insert directories... " if ($opts{v} or $opts{d});
$dbh->begin_work();

# Add all other entries
foreach $key (keys %lookup) {
	# Only run on directories
	if (not exists($lookup{$key}{'type'}) or $lookup{$key}{'type'} != FT_DIR) {
		next;
	}

	my $node = $key;
	my $path = "";

	while (not $lookup{$node}{'parent'} == 0) {
		$path = $lookup{$node}{'name'} . "/$path";
		$node = $lookup{$node}{'parent'};
	}

	$path = VOLUME . '/' . $path;
	$sth->execute($path, $path);
	$sth2->execute($path);

	$lookup{$key}{'dir_no'} = $sth2->fetchrow_array();
}

$dbh->commit();

if ($dbh->err) {
	print STDERR "Failed to commit directories to database.\n";
	&db_disconnect("\n"); # Improve readability of logfile
	exit $dbh->err;
}

print STDERR "Done\n" if ($opts{v} or $opts{d});

$sth = undef;
$sth2 = undef;

# Generate tuples to COPY into FILES_TABLE
print STDERR "Insert files... " if ($opts{v} or $opts{d});
$dbh->do("COPY " . FILES_TABLE . "(backup_no, end_backup_no, dir_no, fname,
	type, mod_date, blk_offset, size, blocks) FROM STDIN");

# Add "@@@:" entry manually
if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
	$dbh->pg_putline(
		"$backup_no\t" .
		END_BACKUP_NO . "\t" .
		"$volume_dir_no\t" .
		VOLUME . "\t" .
		FT_CONTAINER . "\t" .
		"0\t" .
		"0\t" .
		"0\t" .
		"0\n");
} else {
	$dbh->pg_putcopydata(
		"$backup_no\t" .
		END_BACKUP_NO . "\t" .
		"$volume_dir_no\t" .
		VOLUME . "\t" .
		FT_CONTAINER . "\t" .
		"0\t" .
		"0\t" .
		"0\t" .
		"0\n");
}

# Add all other entries
foreach $key (keys %lookup) {
	if ($key == 0 or not exists($lookup{$key}{'parent'}) or $lookup{$key}{'parent'} == 0) {
		next;
	}

	if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
		$dbh->pg_putline(
			"$backup_no\t" .
			END_BACKUP_NO . "\t" .
			"$lookup{$lookup{$key}{'parent'}}{'dir_no'}\t" .
			"$lookup{$key}{'name'}\t" .
			"$lookup{$key}{'type'}\t" .
			"$lookup{$key}{'mod_date'}\t" .
			"$lookup{$key}{'block_offset'}\t" .
			"0\t" .
			"0\n");
	} else {
		$dbh->pg_putcopydata(
			"$backup_no\t" .
			END_BACKUP_NO . "\t" .
			"$lookup{$lookup{$key}{'parent'}}{'dir_no'}\t" .
			"$lookup{$key}{'name'}\t" .
			"$lookup{$key}{'type'}\t" .
			"$lookup{$key}{'mod_date'}\t" .
			"$lookup{$key}{'block_offset'}\t" .
			"0\t" .
			"0\n");
	}
}

print STDERR "Done\n" if ($opts{v} or $opts{d});
if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
	$dbh->pg_endcopy();
} else {
	$dbh->pg_putcopyend();
}

# If there was an error exiting COPY mode, the transaction will fail to commit;
# this results in a backup which has no file names in the database, making
# selective restores impossible. Attempt to insert the files one at a time
# using prepared statements to get as many file names as possible into the
# database and warn the user that some files may be unavailable for selective
# restore. Note that this is MUCH (~60 times) slower and should be used only as
# a last resort.
if ($dbh->err) {
	print STDERR "Warning: Backup $backup_no (or some files in it) may be ",
		"ineligible for selective restore.\n";

	# The CentOS 5 version of the DBD::Pg module does not reset the internal
	# COPY flag if pg_endcopy() fails. The database connection needs to be
	# reset to issue further commands.
	if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
		&db_disconnect;
		&db_connect;
	}

	$sth = $dbh->prepare("INSERT INTO " . FILES_TABLE . " (backup_no,
		end_backup_no, dir_no, fname, type, mod_date, blk_offset, size, blocks)
		VALUES(?, ?, ?, ?, ?, ?, ?, 0, 0)");

	# Add "@@@:" entry manually
	$sth->execute($backup_no, END_BACKUP_NO, $volume_dir_no, VOLUME, FT_CONTAINER, 0, 0);

	foreach $key (keys %lookup) {
		if ($key == 0 or not exists($lookup{$key}{'parent'}) or $lookup{$key}{'parent'} == 0) {
			next;
		}

		$sth->execute($backup_no, END_BACKUP_NO,
			$lookup{$lookup{$key}{'parent'}}{'dir_no'}, $lookup{$key}{'name'},
			$lookup{$key}{'type'}, $lookup{$key}{'mod_date'},
			$lookup{$key}{'block_offset'});
	}

	&db_disconnect("\n"); # Improve readability of logfile
	exit 254; # Let caller know something went wrong and to flag this backup with a warning.
}

&db_disconnect("\n"); # Improve readability of logfile
